#pragma once

#include <utility>
#include <vector>
#include <exception>

using namespace std;

enum phase { WANTS_TO_REPLY, WAITING, READY, TIRED };

class testcase_adversary_base {
  public:
    virtual ~testcase_adversary_base() {}

    virtual void send(int instance, int G, int P, int B) { send(instance, P, B); }
    virtual void send(int instance, int P, int B)        { send(instance, B); }
    virtual void send(int instance, int B)               { throw logic_error("not implemented"); }
    virtual phase getPhase(int instance) = 0;
    virtual int getReply(int instance) = 0;
    virtual void finish(int instance) = 0;
};

class testcase_adversary : public testcase_adversary_base {
  public:
    virtual int write(int G, int P , int B) { return write(P, B); }
    virtual int write(int P, int B)         { return write(B); }
    virtual int write(int B)                { throw logic_error("not implemented"); }

    virtual void send(int instance, int G, int P, int B) {
      assert(instance == 0);
      assert(my_phase == READY);
      reply = write(G, P, B);
      my_phase = WANTS_TO_REPLY;
    }

    virtual phase getPhase(int instance) {
      assert(instance == 0);
      return my_phase;
    }

    virtual int getReply(int instance) {
      assert(instance == 0);
      assert(my_phase == WANTS_TO_REPLY);
      my_phase = READY;
      return reply;
    }

    virtual void finish(int instance) {
      assert(instance == 0);
      my_phase = TIRED;
    }

  protected:
    int reply;
    phase my_phase = READY;
};

class testcase_adversary_parallel_independent : public testcase_adversary {
  public:
    // We take ownership of these pointers
    testcase_adversary_parallel_independent(vector<testcase_adversary_base *> adversaries) :
      adversaries(adversaries) {}
    virtual ~testcase_adversary_parallel_independent() {
      for(auto adv : adversaries) delete adv;
    }

    #define wraps(f, ...) {\
      assert(0 <= instance and instance < adversaries.size()); \
      return adversaries[instance]->f(0, ##__VA_ARGS__); \
    }

    virtual void send(int instance, int G, int P, int B) wraps(send, G, P, B);
    virtual void send(int instance, int P, int B) wraps(send, P, B);
    virtual void send(int instance, int B) wraps(send, B);
    virtual phase getPhase(int instance) wraps(getPhase);
    virtual int getReply(int instance) wraps(getReply);
    virtual void finish(int instance) wraps(finish);

    #undef wraps

  protected:
    vector<testcase_adversary_base *> adversaries;
};

template<typename T, typename... Ps>
testcase_adversary_parallel_independent *bundle_adversaries(int num_instances, Ps... params) {
  vector<testcase_adversary_base *> adversaries;
  for(int i = 0; i < num_instances; ++i) adversaries.push_back(new T(params...));
  return new testcase_adversary_parallel_independent(adversaries);
}

class record_output_testcase_adversary : public testcase_adversary_base {
  public:
    // We take ownership of this pointer
    record_output_testcase_adversary(testcase_adversary_base *adv) : adv(adv) {}
    virtual ~record_output_testcase_adversary() { delete adv; }

    #define wraps(f, ...) {\
      return adv->f(__VA_ARGS__); \
    }

    virtual void send(int instance, int G, int P, int B) wraps(send, instance, G, P, B);
    virtual void send(int instance, int P, int B) wraps(send, instance, P, B);
    virtual void send(int instance, int B) wraps(send, instance, B);
    virtual phase getPhase(int instance) wraps(getPhase, instance);
    virtual void finish(int instance) wraps(finish, instance);

    #undef wraps

    virtual int getReply(int instance) {
      if(outputs.size() <= instance) outputs.resize(instance + 1);
      int r = adv->getReply(instance);
      outputs[instance].push_back(r);
      return r;
    }

    virtual vector<int> getOutput(int instance) {
      assert(getPhase(instance) == TIRED);
      if(outputs.size() <= instance) outputs.resize(instance + 1);
      return outputs[instance];
    }

  protected:
    testcase_adversary_base *adv;
    vector<vector<int>> outputs;
};

template<typename T, typename... Ps>
record_output_testcase_adversary *make_record_output_adv(Ps... params) {
  testcase_adversary_base *adv = new T(params...);
  return new record_output_testcase_adversary(adv);
}

class testcase_adversary_parallel : public testcase_adversary_base {
  public:
    testcase_adversary_parallel(int num_instances) : replies(num_instances, -1), phases(num_instances, READY) {}
    virtual ~testcase_adversary_parallel() {}

    virtual int getReply(int instance) {
      assert(0 <= instance and instance < phases.size());
      assert(phases[instance] == WANTS_TO_REPLY);
      assert(instance < replies.size());
      phases[instance] = READY;
      return replies[instance];
    }

    virtual phase getPhase(int instance) {
      assert(0 <= instance and instance < phases.size());
      return phases[instance];
    }

    virtual void send(int instance, int B) {
      assert(0 <= instance and instance < phases.size());
      assert(phases[instance] == READY);
      phases[instance] = WAITING;
      doSend(instance, B);
    }

    virtual void finish(int instance) {
      assert(0 <= instance and instance < phases.size());
      assert(phases[instance] != TIRED);
      phases[instance] = TIRED;
    }

  protected:
    void reply(int instance, int bit) {
      assert(0 <= instance and instance < phases.size());
      assert(phases[instance] == WAITING);
      replies[instance] = bit;
      phases[instance] = WANTS_TO_REPLY;
    }

    virtual void doSend(int instance, int bit) = 0;

    vector<int> replies;
    vector<phase> phases;
};

class testcase_adversary_unison : public testcase_adversary_parallel {
  public:
    testcase_adversary_unison(int num_instances) :
      testcase_adversary_parallel(num_instances), wants_to_send(num_instances, 0), num_sent(0), num_tired(0) {}
    virtual ~testcase_adversary_unison() {}

    virtual void finish(int instance) {
      testcase_adversary_parallel::finish(instance);
      ++num_tired;
      readyWhenYouAre();
    }

  protected:
    vector<int> wants_to_send;
    int num_sent;
    int num_tired;

    virtual void doSend(int instance, int B) {
      wants_to_send[instance] = B;
      ++num_sent;
      readyWhenYouAre();
    }

    virtual void readyWhenYouAre() {
      if(num_sent + num_tired == wants_to_send.size()) {
        vector<int> to_send = step(wants_to_send);
        assert(to_send.size() == wants_to_send.size());
        for(int i = 0; i < to_send.size(); ++i)
          if(phases[i] != TIRED) reply(i, to_send[i]);
        num_sent = 0;
      }
    }

    virtual vector<int> step(vector<int> wants_to_send) = 0;
};

class target_testcase_adversary : public testcase_adversary {
  public:
    int write(int G, int P, int B) {
      int t = target(P, B);
      if (B != t && g >= G) {
        g = 0;
        written(P, B, t, t);
        return t;
      } else {
        g++;
        written(P, B, t, B);
        return B;
      }
    }
  protected:
    virtual int target(int P, int B) {
      return target(P);
    }
    virtual int target(int P) {
      return 0;
    }
    virtual void written(int P, int B, int T, int R) {
      written(P, R);
    }
    virtual void written(int P, int R) {
      written(R);
    }
    virtual void written(int R) {}
  private:
    int g = 2;
};

vector<vector<char>> sent;

class adversary {
  public:
    virtual ~adversary() {}

    virtual pair<int, int> testcase(int T, int t, int MAX_N, int G) {
      return testcase(T, t, MAX_N);
    }
    virtual pair<int, int> testcase(int T, int t, int MAX_N) {
      return {MAX_N, rand() % MAX_N + 1};
    }
    virtual void send(int instance, int G, int P, int B) {
      testcaseAdversary->send(instance, G, P, B);
    }

    virtual phase getPhase(int instance) {
      return testcaseAdversary->getPhase(instance);
    }

    virtual int getReply(int instance) {
      int r = testcaseAdversary->getReply(instance);
      if(instance >= curr.size()) curr.resize(instance + 1);
      curr[instance].push_back(r);
      return r;
    }

    virtual void testcaseEnd() {
      if (testcaseAdversary) {
        delete testcaseAdversary;
        testcaseAdversary = 0;

        for(auto x : curr) sent.push_back(x);
        curr.clear();
      }
    }

    virtual void newRun() {}

    virtual void finish(int instance) {
      assert(testcaseAdversary != 0);
      testcaseAdversary->finish(instance);
    }

  protected:
    testcase_adversary_base *testcaseAdversary = 0;
    vector<vector<char>> curr;
};